import numpy as np
import pandas as pd


def vN_entropy(M, SUB_M_SIZE_FIX, CHRSPLIT, PHI_MAX, phi, BIN_TABLE):
    S = []
    BIN_TABLE_NEW = []
    N = M.shape[0]

    if np.isnan(SUB_M_SIZE_FIX):
        SUB_M_SIZE = np.round(N / CHRSPLIT)
    else:
        SUB_M_SIZE = SUB_M_SIZE_FIX

    SUB_M_SIZE = int(SUB_M_SIZE)
    WN = 1 + np.floor((N - SUB_M_SIZE) / phi)
    while WN > PHI_MAX:
        phi = phi + 1
        WN = 1 + np.floor((N - SUB_M_SIZE) / phi)
    WN = int(WN)

    R1 = np.array(range(0, N, phi))
    R1 = R1[0:WN]
    R2 = R1 + SUB_M_SIZE - 1
    R2 = R2[0:WN]
    R = np.stack((R1, R2), axis=1).astype(int)

    M = M.astype(float)  # cannot fill with nan otherwise
    M[M == 0] = np.nan
    M = np.log(M)
    # with open("outputPY.txt", "w") as FN:
    for rr in range(0, WN):
        # Python slicing is EXCLUSIVE of the stop index

        m = M[R[rr, 0] : R[rr, 1] + 1, R[rr, 0] : R[rr, 1] + 1].copy()

        # A view is a new array object that looks like a separate array but
        # shares the same data in memory as the original array. If you modify m,
        # you’re also modifying M in-place.
        # np.savetxt("matrix_python.isolated.txt", m, fmt="%.15g", delimiter="\t")

        # see GIT: NormM=0 and ChrY
        # print(np.any(np.sum((np.isnan(m) | (m == 0)), axis=1) >= (m.shape[0] - 1)))

        if np.all(np.isnan(m) | (m == 0)):
            ENT = np.nan

        else:
            # if any row/column is nan/0, ignore this column (0 comes from log(1)!)
            mask = np.sum((np.isnan(m) | (m == 0)), axis=1) < (SUB_M_SIZE)
            m = m[mask, :]
            m = m[:, mask]

            m[np.isnan(m)] = np.nanmin(m.flatten())
            P = np.corrcoef(m, rowvar=False)
            # problems with corrcoef and const. columns.
            # m = np.array([[12, 3, 2], [2, 3, 11], [1, 3, 7]])
            # m = np.array([[12, 0.3, 2], [2, 0.3, 11], [1, 0.3, 7]])
            # print(np.std(m, axis=0, ddof=1))
            # NumPy computes population std (N in denom), matlab the sample std ((N-1) in denom)!
            # I need np.std(m, axis=0, ddof=1)
            SDs = np.std(m, axis=0, ddof=1) < np.finfo(float).eps
            P[SDs, :] = 0
            P[:, SDs] = 0
            P[np.isnan(P)] = 0  #
            np.fill_diagonal(P, 1.0)
            rho = P / P.shape[0]
            # Compute the eigenvalues of a complex Hermitian or real symmetric matrix.
            vals = np.linalg.eigvalsh(rho)
            vals = vals[vals > np.finfo(float).eps]
            vals = np.real(vals)
            ENT = -np.sum(vals * np.log(vals))

        S.append(ENT)
        # print(f"Step {rr}: S = {ENT:.7f}")  # , file=FN)

        BIN_TABLE_NEW.append(
            [
                BIN_TABLE.iloc[R[rr, 0], BIN_TABLE.columns.get_loc("binNr")],
                BIN_TABLE.iloc[R[rr, 1], BIN_TABLE.columns.get_loc("binNr")],
                BIN_TABLE.iloc[R[rr, 0], BIN_TABLE.columns.get_loc("start")],
                BIN_TABLE.iloc[R[rr, 1], BIN_TABLE.columns.get_loc("end")],
            ]
        )

    S = np.array(S, dtype=np.float64)

    BIN_TABLE_NEW = pd.DataFrame(
        BIN_TABLE_NEW, columns=["binNr1", "binNr2", "start", "end"]
    )
    return S, SUB_M_SIZE, WN, phi, BIN_TABLE_NEW



import os
import argparse
import logging
from ENT3C.core import get_similarity, get_entropy
from ENT3C.core.utils import check_config


def run_get_entropy(config_file):
    get_entropy(config_file)


def run_get_similarity(config_file):
    get_similarity(config_file)


def run_all(config_file):
    get_entropy(config_file)
    get_similarity(config_file)


def main():
    # top-level parser: main ArgumentParser instance

    parser = argparse.ArgumentParser(
        prog="ENT3C",
        description="Compute the similarity between genomic contact matrices.",
        epilog="For more info, visit https://github.com/X3N1A/ENT3C",
    )

    subparsers = parser.add_subparsers(
        dest="command", required=True
    )  # subcommand stored in args.command

    parser_get_entropy = subparsers.add_parser("get_entropy")
    parser_get_entropy.add_argument(
        "--config",
        required=True,
        help="Path to json configuration file. Generates an entropy output file.",
    )

    parser_get_similarity = subparsers.add_parser("get_similarity")
    parser_get_similarity.add_argument(
        "--config",
        required=True,
        help="Path to json configuration file. Generates a similarity output file from an entropy input file <entropy_out_FN>.",
    )
    # parser_get_similarity.add_argument(
    #    "--entropy_out_FN", required=True, help="Path to entropy table output file."
    # )

    parser_run_all = subparsers.add_parser("run_all")
    parser_run_all.add_argument(
        "--config",
        required=True,
        help="Path to json configuration file. Generates both entropy and similarity output files.",
    )

    args = parser.parse_args()

    if not args.command:
        parser.print_help()

    config_file = os.path.abspath(args.config)

    print(config_file)
    print(args)

    (
        config_df,
        SUB_M_SIZE_FIX,
        PHI_MAX,
        CHRSPLIT,
        phi,
        NormM,
        CHROMOSOMES,
        RESOLUTIONS,
        BR,
        FNs,
        OUT_DIR,
        OUT_PREFIX,
        entropy_out_FN,
        similarity_out_FN,
    ) = check_config(config_file)

    LOG_FN = f"{OUT_DIR}/{OUT_PREFIX}_logfile.log"

    logging.basicConfig(
        filename=LOG_FN,
        level=logging.INFO,
        format="%(asctime)s - %(levelname)s - %(message)s",
    )

    logging.info(f"{args}")
    logging.info(f"Inputs: {', '.join(config_df['NAME'])}")
    logging.info(
        "Apply cooler weights? no."
        if config_df["NormM"][0] == 0
        else f"Apply cooler weights? yes. Name in cooler:{', '.join(config_df['WEIGHTS_NAME'][0])}"
    )
    logging.info(f"CHRSPLIT: {CHRSPLIT}")
    logging.info(f"Sub matrix size PHI: {SUB_M_SIZE_FIX}")
    logging.info(f"Window shift phi: {phi}")
    logging.info(f"Chromosomes: {CHROMOSOMES}")
    logging.info(f"Resolutions: {RESOLUTIONS}")

    if args.command == "get_entropy":
        config_file = os.path.abspath(args.config)
        print("Using config file:", config_file)
        get_entropy(config_file)
        print(f"entropy table in: {entropy_out_FN}")

    if args.command == "get_similarity":
        config_file = os.path.abspath(args.config)
        print("Using config file:", config_file)
        print("Similarity accordinf to entropy file:", {entropy_out_FN})
        get_similarity(config_file)
        print(f"similarity table in: {similarity_out_FN}")

    if args.command == "run_all":
        config_file = os.path.abspath(args.config)

        print("Using config file:", config_file)
        get_entropy(config_file)
        get_similarity(config_file)
        print(f"entropy table in: {entropy_out_FN}")
        print(f"similarity table in: {similarity_out_FN}")

    LOG_FN = f"{OUT_DIR}/{OUT_PREFIX}_logfile.log"
    logging.info(args)
    print(f"log file in: {LOG_FN}")


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd

from ENT3C.core import utils
from ENT3C.core.vN_entropy import vN_entropy


def get_entropy(config_file):
    (
        config_df,
        SUB_M_SIZE_FIX,
        PHI_MAX,
        CHRSPLIT,
        phi,
        NormM,
        CHROMOSOMES,
        RESOLUTIONS,
        BR,
        FNs,
        OUT_DIR,
        OUT_PREFIX,
        entropy_out_FN,
        similarity_out_FN,
    ) = utils.check_config(config_file)

    print(f"Inputs: {', '.join(config_df['NAME'])}")
    print(
        "Apply cooler weights? no."
        if config_df["NormM"][0] == 0
        else f"Apply cooler weights? yes. Name in cooler:{config_df['WEIGHTS_NAME'][0]}"
    )
    print(f"CHRSPLIT: {CHRSPLIT}")
    print(f"Sub matrix size PHI: {SUB_M_SIZE_FIX}")
    print(f"Window shift phi: {phi}")
    print(f"Chromosomes: {CHROMOSOMES}")
    print(f"Resolutions: {RESOLUTIONS}")

    print(f"Output Directory: {config_df['OUT_DIR'][0]}")
    print(f"Output Prefix: {config_df['OUT_PREFIX'][0]}")

    ENT3C_OUT = []
    print(f"Generating new file: {entropy_out_FN}")
    for Resolution in RESOLUTIONS:
        Resolution = int(Resolution)
        for ChrNr in CHROMOSOMES:
            ######## locate common empty bins ########
            EXCLUDE = set()  # use set() to keep unique values

            for FN in FNs:
                BIN_TABLE, M = utils.load_cooler(
                    FN,
                    ChrNr,
                    Resolution,
                    config_df["NormM"][0],
                    config_df["WEIGHTS_NAME"][0],
                )

                if NormM == 0:
                    EXCLUDE.update(BIN_TABLE[BIN_TABLE["CONTACT"].isna()]["binNr"])
                elif NormM == 1:
                    EXCLUDE.update(
                        BIN_TABLE[
                            BIN_TABLE["CONTACT"].isna() | BIN_TABLE["weight"].isna()
                        ]["binNr"]
                    )

            #########################################
            # N = len(FNs)
            # cols = int(np.ceil(np.sqrt(N)))
            # rows = int(np.ceil(N / cols))
            # fig, axs = plt.subplots(rows, cols, figsize=(cols * 4, rows * 4))
            # axs = axs.flatten()
            # cmap = mpl.colormaps.get_cmap("magma")
            # cmap.set_bad(color="black")
            # CELL = config[(config["DATA_PATH"] + "/" + config["FILE"] == FN)]["NAME"].iloc[0]
            # axs[i].set_title(FN)
            # i = i + 1
            # plt.tight_layout()
            # plt.show()
            ######### compute entropies ##############
            EXCLUDE = set(sorted(EXCLUDE))

            for f in range(0, len(FNs)):
                BIN_TABLE, M = utils.load_cooler(
                    FNs[f],
                    ChrNr,
                    Resolution,
                    config_df["NormM"][0],
                    config_df["WEIGHTS_NAME"][0],
                )

                INCLUDE = set(range(0, M.shape[0]))
                INCLUDE = INCLUDE - EXCLUDE

                BIN_TABLE = BIN_TABLE.iloc[list(INCLUDE), :].reset_index(drop=True)
                M = M[np.meshgrid(sorted(INCLUDE), sorted(INCLUDE), indexing="ij")]

                S, SUB_M_SIZE, WN, phi, BIN_TABLE_NEW = vN_entropy(
                    M, SUB_M_SIZE_FIX, CHRSPLIT, PHI_MAX, phi, BIN_TABLE
                )

                N = len(S)
                DICT = {
                    "Name": [config_df["NAME"].iloc[f]] * N,
                    "ChrNr": [ChrNr] * N,
                    "Resolution": [Resolution] * N,
                    "n": [SUB_M_SIZE] * N,
                    "PHI": [WN] * N,
                    "phi": [phi] * N,
                    "binNrStart": BIN_TABLE_NEW.iloc[:, 0],
                    "binNrEnd": BIN_TABLE_NEW.iloc[:, 1],
                    "START": BIN_TABLE_NEW["start"],
                    "END": BIN_TABLE_NEW["end"],
                    "S": S,
                }
                ENT3C_OUT.append(pd.DataFrame(DICT))  # stores DFs in list

    ENT3C_OUT = pd.concat(ENT3C_OUT)
    ENT3C_OUT["ChrNr"] = ENT3C_OUT["ChrNr"].astype(str)

    ENT3C_OUT.to_csv(
        entropy_out_FN,
        index=False,
        sep="\t",
        na_rep="NaN",
    )

    print("ENT3C Output table:")
    print(ENT3C_OUT)

    return ENT3C_OUT



import numpy as np
import matplotlib.pyplot as plt
import itertools
import pandas as pd
from ENT3C.core import utils


def get_similarity(config_file):
    (
        config_df,
        SUB_M_SIZE_FIX,
        PHI_MAX,
        CHRSPLIT,
        phi,
        NormM,
        CHROMOSOMES,
        RESOLUTIONS,
        BR,
        FNs,
        OUT_DIR,
        OUT_PREFIX,
        entropy_out_FN,
        similarity_out_FN,
    ) = utils.check_config(config_file)

    ENT3C_OUT = pd.read_csv(f"{entropy_out_FN}", sep="\t")
    ENT3C_OUT["ChrNr"] = ENT3C_OUT["ChrNr"].astype(str)

    SAMPLES = set(ENT3C_OUT["Name"])
    if any("BR" in element for element in SAMPLES):
        Biological_replicates = True
    else:
        Biological_replicates = False

    Similarity = pd.DataFrame(
        {
            "Resolution": pd.Series(dtype="int"),
            "ChrNr": pd.Series(dtype="str"),
            "Sample1": pd.Series(dtype="str"),
            "Sample2": pd.Series(dtype="str"),
            "Q": pd.Series(dtype="float"),
        }
    )

    cols = int(np.ceil(np.sqrt(len(CHROMOSOMES))))
    rows = int(np.ceil(len(CHROMOSOMES) / cols))
    if len(SAMPLES) > 1:
        comparisons = list(itertools.combinations(SAMPLES, 2))
        for Resolution in RESOLUTIONS:
            fig, axs = plt.subplots(
                nrows=rows, ncols=cols, figsize=(cols * 8, rows * 4)
            )  # Width, height in inches
            fig.suptitle(f"{Resolution / 1e3}kb")

            if isinstance(axs, np.ndarray):
                axs = axs.flatten()
            else:
                axs = np.array([axs])

            for i, ChrNr in enumerate(CHROMOSOMES):
                ax = axs[i]
                ax.set_title(f"Chr{ChrNr}")

                plotted = set()

                for comp in comparisons:
                    S1 = ENT3C_OUT[
                        (ENT3C_OUT["Name"] == comp[0])
                        & (ENT3C_OUT["ChrNr"] == ChrNr)
                        & (ENT3C_OUT["Resolution"] == Resolution)
                    ]["S"]
                    S1 = np.array(S1, dtype=np.float64)
                    S2 = ENT3C_OUT[
                        (ENT3C_OUT["Name"] == comp[1])
                        & (ENT3C_OUT["ChrNr"] == ChrNr)
                        & (ENT3C_OUT["Resolution"] == Resolution)
                    ]["S"]
                    S2 = np.array(S2, dtype=np.float64)

                    non_nan_idx = ~np.isnan(S1) & ~np.isnan(S2)
                    Q = np.corrcoef(S1[non_nan_idx], S2[non_nan_idx])[0, 1]

                    new_row = pd.DataFrame(
                        {
                            "Resolution": [Resolution],
                            "ChrNr": [str(ChrNr)],
                            "Sample1": [comp[0]],
                            "Sample2": [comp[1]],
                            "Q": [Q],
                        }
                    )

                    Similarity = pd.concat([Similarity, new_row], ignore_index=True)

                    if comp[0] not in plotted:
                        ax.plot(S1, label=f"{comp[0]}")
                        plotted.add(comp[0])
                        i += 1
                    if comp[1] not in plotted:
                        ax.plot(S2, label=f"{comp[1]}")
                        plotted.add(comp[1])
                        i += 1

                if Biological_replicates:
                    Q_BR = Similarity[
                        (Similarity["ChrNr"] == ChrNr)
                        & (Similarity["Resolution"] == Resolution)
                        & (
                            Similarity["Sample1"].str.split("_").str[0]
                            == Similarity["Sample2"].str.split("_").str[0]
                        )
                    ]["Q"].mean()

                    Q_NR = Similarity[
                        (Similarity["ChrNr"] == ChrNr)
                        & (Similarity["Resolution"] == Resolution)
                        & (
                            Similarity["Sample1"].str.split("_").str[0]
                            != Similarity["Sample2"].str.split("_").str[0]
                        )
                    ]["Q"].mean()

                    title_str = (
                        rf"Chr{ChrNr}"
                        + "\n"
                        + rf"$\overline{{Q}}_{{BR}} = {Q_BR:.2f}$ "
                        + rf"$\overline{{Q}}_{{NR}} = {Q_NR:.2f}$"
                    )
                    ax.set_title(title_str, fontsize=25)
                else:
                    Q_NR = Similarity[
                        (Similarity["ChrNr"] == ChrNr)
                        & (Similarity["Resolution"] == Resolution)
                        & (
                            Similarity["Sample1"].str.split("_").str[0]
                            != Similarity["Sample2"].str.split("_").str[0]
                        )
                    ]["Q"].mean()
                    title_str = rf"Chr{ChrNr}" + rf"$\overline{{Q}} = {Q_NR:.2f}$"
                    ax.set_title(title_str, fontsize=25)

                ax.tick_params(labelsize=15)
                ax.title.set_fontsize(15)
                ax.xaxis.label.set_fontsize(15)
                ax.yaxis.label.set_fontsize(15)
                ax.autoscale(enable=True, axis="both", tight=True)

            if ChrNr == CHROMOSOMES[-1]:
                plotted = [p.replace("_", " ") for p in plotted]
                ax.legend(
                    plotted,
                    loc="upper right",
                    bbox_to_anchor=(1.2, 1),
                    borderaxespad=0.0,
                )
            for i in range(len(CHROMOSOMES), len(axs)):
                axs[i].axis("off")

            plt.tight_layout()
            plt.savefig(
                f"{OUT_DIR}/{OUT_PREFIX}_{Resolution}_ENT3C_signals.svg",
                bbox_inches="tight",
            )  # bbox_inches trims whitespace
    print("Output similarity table:")
    print(Similarity)

    Similarity.to_csv(
        similarity_out_FN,
        index=False,
        sep="\t",
    )

    return Similarity



import numpy as np
import cooler as cl
import pandas as pd
import os
import logging


def check_config(config_path):
    with open(config_path, "r") as file:
        config_df = pd.read_json(file)

    FILES = config_df["FILES"]
    # pandas aligns the rows by index.
    # FILES[::2] has indices 0,2,4,.. and FILES[1::2] has 1,2,3
    config_files = pd.DataFrame(
        {
            "FILE": FILES[::2].reset_index(drop=True),
            "NAME": FILES[1::2].reset_index(drop=True),
        }
    )

    del config_df["FILES"]

    config_df = config_df[::2].reset_index(drop=True)
    config_df = pd.concat([config_df, config_files], axis=1)

    if "BR" in config_df["NAME"].astype(str):
        BR = True
    else:
        BR = False

    config_df["OUT_DIR"] = os.path.join(config_df["OUT_DIR"][0], "PYTHON")

    if pd.isna(config_df["OUT_PREFIX"]).all():
        config_df["OUT_PREFIX"] = "%dkb" % (config_df["Resolution"][0] / 1e3)
    # print("%dkb" % (config_df["Resolution"][0]/1e3))

    if not os.path.exists(config_df["OUT_DIR"][0]):
        os.makedirs(config_df["OUT_DIR"][0])

    config_df["ChrNr"] = config_df["ChrNr"].astype(str)

    if "," in config_df["ChrNr"][0]:
        CHROMOSOMES = config_df["ChrNr"][0].split(",")
    else:
        CHROMOSOMES = [config_df["ChrNr"][0]]

    config_df["Resolution"] = config_df["Resolution"].astype(str)

    if "," in str(config_df["Resolution"][0]):
        RESOLUTIONS = [int(float(r)) for r in config_df["Resolution"][0].split(",")]
    else:
        RESOLUTIONS = [int(config_df["Resolution"][0])]

    FNs = config_df["DATA_PATH"] + "/" + config_df["FILE"]
    FNs = FNs.tolist()

    SUB_M_SIZE_FIX = config_df["SUB_M_SIZE_FIX"].unique()
    PHI_MAX = config_df["PHI_MAX"].unique()
    CHRSPLIT = config_df["CHRSPLIT"].unique()
    OUT_DIR = config_df["OUT_DIR"].unique()[0]
    OUT_PREFIX = config_df["OUT_PREFIX"].unique()[0]

    os.makedirs(OUT_DIR, exist_ok=True)
    phi = config_df["phi"].unique()
    NormM = config_df["NormM"].unique()
    if SUB_M_SIZE_FIX.size > 1:
        raise ValueError("SUB_M_SIZE_FIX defined multiple times in config file")
    elif PHI_MAX.size > 1:
        raise ValueError("PHI_MAX defined multiple times in config file")
    elif CHRSPLIT.size > 1:
        raise ValueError("CHRSPLIT defined multiple times in config file")
    elif phi.size > 1:
        raise ValueError("phi defined multiple times in config file")
    else:
        SUB_M_SIZE_FIX = SUB_M_SIZE_FIX[0]
        PHI_MAX = PHI_MAX[0]
        CHRSPLIT = CHRSPLIT[0]
        phi = phi[0]
        NormM = NormM[0]

    entropy_out_FN = f"{OUT_DIR}/{OUT_PREFIX}_ENT3C_OUT.csv"
    similarity_out_FN = f"{OUT_DIR}/{OUT_PREFIX}_ENT3C_similarity.csv"

    LOG_FN = f"{OUT_DIR}/{OUT_PREFIX}_logfile.log"

    logging.basicConfig(
        filename=LOG_FN,
        level=logging.INFO,
        format="%(asctime)s - %(levelname)s - %(message)s",
    )

    logging.info(f"Inputs: {', '.join(config_df['NAME'])}")
    logging.info(
        "Apply cooler weights? no."
        if config_df["NormM"][0] == 0
        else f"Apply cooler weights? yes. Name in cooler:{', '.join(config_df['WEIGHTS_NAME'][0])}"
    )
    logging.info(f"CHRSPLIT: {CHRSPLIT}")
    logging.info(f"Sub matrix size PHI: {SUB_M_SIZE_FIX}")
    logging.info(f"Window shift phi: {phi}")
    logging.info(f"Chromosomes: {CHROMOSOMES}")
    logging.info(f"Resolutions: {RESOLUTIONS}")

    return (
        config_df,
        SUB_M_SIZE_FIX,
        PHI_MAX,
        CHRSPLIT,
        phi,
        NormM,
        CHROMOSOMES,
        RESOLUTIONS,
        BR,
        FNs,
        OUT_DIR,
        OUT_PREFIX,
        entropy_out_FN,
        similarity_out_FN,
    )


def load_cooler(FN, ChrNr, Resolution, NormM, weights_name):
    if "mcool" in FN:
        clr = cl.Cooler("%s::resolutions/%d" % (FN, Resolution))
    else:
        clr = cl.Cooler(FN)

    R = clr.binsize

    assert R == Resolution, f"Expected {Resolution} resolution in {FN}, found {R}"

    # Table selectors (chroms, bins, pixels)
    BIN_TABLE = clr.bins().fetch("chr" + str(ChrNr))

    BIN_TABLE = BIN_TABLE.reset_index()
    BIN_TABLE.rename(
        columns={
            "index": "BINS_ALL",
            "chrom": "chrs",
            "start": "START",
            "end": "END",
            "weight": "weights",
        }
    )
    BIN_TABLE["binNr"] = range(0, len(BIN_TABLE))
    if NormM:
        M = clr.matrix(balance=True).fetch("chr" + str(ChrNr))
    else:
        M = clr.matrix(balance=False).fetch("chr" + str(ChrNr))

    INCLUDE = np.where(~np.all(np.isnan(M), axis=1) & ~np.all(M == 0, axis=1))[0]

    BIN_TABLE["CONTACT"] = np.full((BIN_TABLE.shape[0],), np.nan)

    BIN_TABLE.loc[INCLUDE, "CONTACT"] = 1

    return BIN_TABLE, M


def get_cell_line(sample):
    cell = sample.split("_BR")[0]
    return cell
